/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

#include <stdio.h>
#include <direct.h>
#include <sys/stat.h>
#include <TeAdoSqlServer.h>
#include <sys/stat.h>

#include <TeUtils.h>

#define CHUNKSIZE	240

typedef map<int,TeNode> TeNodeMap;

inline void TESTHR( HRESULT hr )
{
	if( FAILED(hr) ) _com_issue_error( hr );
}

TeSqlServer::TeSqlServer()  
{
    HRESULT hr = CoInitialize(0);
	dbmsName_ = "SqlServerAdo";
    if(FAILED(hr))
    {
        cout << "Can't start COM!? " << endl;
    }
}

TeSqlServer::~TeSqlServer()
{
}

bool 
TeSqlServer::newDatabase(const string& database, const string& user, const string& password, const string& host, const int& port, bool terralibModel)
{
	if ( connect(host,user,password,database,port) == false )
		return false;
	
	if ( terralibModel )
	{
		//create conceptual model
		if(!this->createConceptualModel())
			return false;
	}	
	return true;
}


bool 
TeSqlServer::connect (const string& host, const string& user, const string& password, const string& database, int /* port */)
{
	
	string  connectionString = "Provider=SQLOLEDB.1;FetchSize=300;Password=" + password + ";Persist Security Info=True;User ID=" + user + ";Initial Catalog=" + database + ";Data Source=" + host;

	try
	{
		connection_.CreateInstance(__uuidof(ADODB::Connection));
		HRESULT hr  = connection_->Open (connectionString.c_str(),"","",-1);
		TESTHR( hr );
	}
	catch(_com_error &e)
	{
		isConnected_ = false;
		errorMessage_ = e.Description();
		return false;
	}
	catch(...)
	{
		isConnected_ = false;
		errorMessage_ = "Oppps !";
		return false;
	}
	isConnected_ = true;
	database_ = database;
	host_ = host;
	user_ = user;
	password_ = password;
	return true;
}


bool
TeSqlServer::tableExist(const string& table)
{
	bool	status;
	TeSqlServerPortal *sqlp = (TeSqlServerPortal*)getPortal();
	if(!sqlp)
		return false;

	string exist = "SELECT name FROM sysobjects WHERE UPPER(name) = UPPER('" + table + "') ";
	exist+=" and UPPER(xtype)='U'";
		
		
		
	if(!sqlp->query(exist))
	{
		delete sqlp;
		return false;
	}

	if(sqlp->fetchRow())
		status = true;
	else
		status = false;

	delete sqlp;
	return (status);
}


bool 
TeSqlServer::columnExist(const string& table, const string& column, TeAttribute& attr)
{
	TeSqlServerPortal *sqlp = (TeSqlServerPortal*)getPortal();
	if(!sqlp)
		return false;

	string exist ="select name,xtype,length,xscale  from syscolumns where name='" + column + "' and ";
	exist+=" id in(select id from sysobjects where xtype='U' and name='" + table + "')";

	
	if(!sqlp->query(exist))
	{
		delete sqlp;
		return false;
	}

	if(sqlp->fetchRow())
	{	
		attr.rep_.name_ = column;
		
		int	dataType = atoi(sqlp->getData(1)); 
		int		dataLength = atoi(sqlp->getData(2)); 
		attr.rep_.numChar_ = dataLength;
		if(dataType == 167 && dataLength < 256)
		{
			attr.rep_.type_ = TeSTRING;
		}
		else if (dataType == 167 && dataLength >= 256)
		{
			attr.rep_.type_ = TeBLOB;
		}
		else if (dataType == 34)
		{
			attr.rep_.type_ = TeBLOB;
		}
		else if (dataType == 56)
		{
			attr.rep_.type_ = TeREAL;
		}
		else if (dataType == 62)
		{
			attr.rep_.type_ = TeINT;
		}
		else if (dataType == 167 && dataLength == 0)
		{	
			attr.rep_.type_ = TeCHARACTER;
		}
		else if (dataType == 61)
		{
			attr.rep_.type_ = TeDATETIME;
		}
		else
		{
			attr.rep_.type_ = TeSTRING;
			attr.rep_.numChar_ = dataLength;
		}
		delete sqlp;
		return true;
	}
	delete sqlp;
	return false;
}

bool 
TeSqlServer::getAttributeList(const string& tableName,TeAttributeList& attList)
{
	return(TeDatabase::getAttributeList (tableName, attList));
}


bool
TeSqlServer::createTable(const string& table, TeAttributeList &attr)
{
	short	cont=0;
	string pkeys ="";

	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}

	TeAttributeList::iterator it = attr.begin();
	string tablec;
	tablec = "CREATE TABLE " + table +" (";
	
	while ( it != attr.end())
	{
		if (cont)
			tablec += ", ";
			
		switch ((*it).rep_.type_)
		{
			case TeSTRING:
				if ( (*it).rep_.numChar_ < 256 )
				   tablec += (*it).rep_.name_ + " VARCHAR(" + Te2String((*it).rep_.numChar_) + ")";
				else
				   tablec += (*it).rep_.name_ + " TEXT"; 
			break;
			
			case TeREAL:
				tablec += (*it).rep_.name_ + " DOUBLE PRECISION ";
			break;
			
			case TeINT:
			case TeUNSIGNEDINT:
				tablec += (*it).rep_.name_ + " INT ";
				if ((*it).rep_.isAutoNumber_) 
					tablec += " IDENTITY(1,1) ";
			break;

			case TeDATETIME:
				tablec += (*it).rep_.name_ + " DATETIME ";
			break;

			case TeCHARACTER:
				tablec += (*it).rep_.name_ + " CHAR ";
			break;

			case TeBLOB:
				tablec += (*it).rep_.name_ + " IMAGE ";
			break;

			default:
				tablec += (*it).rep_.name_ + " VARCHAR(255) ";
			break;
		}
		//default value
		if(!((*it).rep_.defaultValue_.empty()))
            tablec += " DEFAULT '" + (*it).rep_.defaultValue_ +"' " ;

		//not null
		if(!((*it).rep_.null_))
			tablec += " NOT NULL ";
		
		// check if column is part of primary key
		if ((*it).rep_.isPrimaryKey_ && (*it).rep_.type_ != TeBLOB )
		{
			if (!pkeys.empty())
				pkeys += ", ";
			pkeys += (*it).rep_.name_;
			tablec += " NOT NULL";
		}

		++it;
		cont++;
	}

	if(!pkeys.empty())
		tablec += ", PRIMARY KEY (" + pkeys + ") ";

	tablec += ")";

	if(!execute(tablec))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table;

		return false;
	}
	return true;
}

bool 
TeSqlServer::alterTable(const string& oldTableName, const string& newTablename)
{
	string sql = "sp_rename '"+ oldTableName +"', '"+ newTablename +"'";
	if(!this->execute(sql))
		return false;

	//update te_layer_table
	sql = " UPDATE te_layer_table ";
	sql += " SET attr_table = '"+ newTablename +"'";
	sql += " WHERE attr_table = '"+ oldTableName +"'";
	execute(sql);

	//update te_tables_relation
	sql = " UPDATE te_tables_relation ";
	sql += " SET external_table_name = '"+ newTablename +"'";
	sql += " WHERE external_table_name = '"+ oldTableName +"'";
	execute(sql);
	
	alterTableInfoInMemory(newTablename, oldTableName);
	return true;
}

bool
TeSqlServer::deleteTable (const string& table)
{
//	int f =	table.find ("te_collection", std::string::npos);
	int f =	table.find ("te_collection");

	if( table=="te_theme" ||
		table=="te_layer" ||
		table=="te_representation" ||
		table=="te_tables_relation" ||
		table=="te_layer_table" ||
		table=="te_color_scheme" ||
		table=="te_raster_metadata" ||
		table=="te_projection" ||
		table=="te_view" ||
		table=="te_chart_params" ||
		table=="te_legend" ||
		table=="te_visual" ||
		f == 0)
	{
		errorMessage_ = "No  possvel deletar tabelas do modelo!";
		return false;
	}

	string del = "DROP TABLE " + table;
	if(!execute(del))
		return false;
	
	return true;
}

bool 
TeSqlServer::addColumn (const string& table, TeAttributeRep &rep)
{
	if(!tableExist(table))
		return false;

	string tab;
	tab = " ALTER TABLE " + table + " ADD  ";
	tab += rep.name_ + "  ";
	
	switch (rep.type_)
	{
		case TeSTRING:
			if ( rep.numChar_ < 256 )
				tab += "VARCHAR(" + Te2String(rep.numChar_) + ") ";
			else
				tab += "TEXT "; 
		break;
			
		case TeREAL:
			tab += "DOUBLE PRECISION";	
			break;
			
		case TeINT:
			tab += "INT";
			break;

		case TeDATETIME:
			tab += "DATETIME";

		case TeCHARACTER:
			tab += "CHAR";
		
		case TeBLOB:
			tab += "IMAGE";
		
		default:
			tab += "VARCHAR(" + Te2String(rep.numChar_) + ") ";
			break;
	}

	alterTableInfoInMemory(table);
	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error inserting a column to table " + table + " !";
		return false;
	}

	return true;
}


bool
TeSqlServer::createRelation (const string& name, const string& table, const string& fieldName, const string& relatedTable, const string& relatedField, bool cascadeDeletion)
{
	
	string relation = "ALTER TABLE " + table + " ADD ";
	relation += " CONSTRAINT " + name;
	relation += " FOREIGN KEY ( " + fieldName + ") "; 
	relation += " REFERENCES " + relatedTable + "(" + relatedField + ")";
		
	if (cascadeDeletion)
		relation += " ON DELETE CASCADE ON UPDATE NO ACTION"; 

	if(!execute(relation))
		return false;

	return true;
}


TeDBRelationType 
TeSqlServer::existRelation(const string& /* tableName */, const string& relName)
{

	TeSqlServerPortal  *sqlp = (TeSqlServerPortal*)getPortal();
	if(!sqlp)
		return TeNoRelation;

	string integ="select name from sysobjects where upper(name)=upper('" + relName + "')";
	integ +=" and upper(xtype)='F'";
	
	if(!sqlp->query(integ))
	{
		delete sqlp;
		return TeNoRelation;
	}
	

	if ( sqlp->fetchRow() )		
	{
		int status = sqlp->getInt("status");
		if ( (status & 14338) == 14338 ) //Cascade Update Delete 
		{
			delete sqlp;
			return TeRICascadeDeletion; 
		}
		if ( (status & 4098) == 4098 ) //Cascade Delete
		{
			delete sqlp;
			return TeRICascadeDeletion;
		}
		delete sqlp;
		return TeRINoCascadeDeletion;
	}

	delete sqlp;
	return TeNoRelation;	
}


TeDatabasePortal*  
TeSqlServer::getPortal ()
{
	TeSqlServerPortal* portal = new TeSqlServerPortal (this);
	return portal;
}

bool 
TeSqlServer::createDatabaseTable()
{
	string create = "CREATE TABLE te_database (";
	create += "db_version		VARCHAR(50) NOT NULL,";
	create += "db_creation		DATETIME,";
	create += "PRIMARY KEY (db_version))";
	return (execute (create));
}

bool 
TeSqlServer::createProjectionTable()
{

	string table;
	table = "CREATE TABLE te_projection (";
	table += "projection_id ";
	table += "INT IDENTITY(1,1) NOT NULL";
	table += ", name ";
	table += "VARCHAR(255) NOT NULL";
	table += ", long0 ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL ";
	table += ", lat0 ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL ";
	table += ", offx ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL ";
	table += ", offy ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL ";
	table += ", stlat1 ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL ";
	table += ", stlat2 ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL ";
	table += ", unit ";
	table += "VARCHAR(100) ";
	table += ", scale ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL ";
	table += ", hemis ";
	table += "NUMERIC(20) ";
	table += ", datum ";
	table += "VARCHAR(100) ";
	table += ", radius ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL ";
	table += ", flattening ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL ";
	table += ", dx ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL ";
	table += ", dy ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL ";
	table += ", dz ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL ";
	
	table += ", PRIMARY KEY (projection_id))";

	if(!execute(table))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_projection!";
		return false;
	}
	
	
	return true;
}



bool 
TeSqlServer::createLayerTable()
{
	string table;
	table = "CREATE TABLE te_layer (";
	table += "layer_id ";
	table += "INT IDENTITY(1,1) NOT NULL";
	table += ", projection_id ";
	table += "INT NOT NULL ";
	table += "CONSTRAINT layer_proj UNIQUE ";
	table += ", name ";
	table += "VARCHAR(255) NOT NULL ";
	table += ", lower_x ";
	table += "DOUBLE PRECISION";
	table += ", lower_y ";
	table += "DOUBLE PRECISION";
	table += ", upper_x ";
	table += "DOUBLE PRECISION";
	table += ", upper_y ";
	table += "DOUBLE PRECISION";
	table += ", initial_time ";
	table += "DATETIME ";
	table += ", final_time ";
	table += "DATETIME ";

	table += ", PRIMARY KEY (layer_id))";

	if(!execute(table))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_layer!";
		return false;
	}

	return true;
}

bool 
TeSqlServer::createLayerTableTable()
{
	string table;
	table = "CREATE TABLE te_layer_table (";
	table += "table_id ";
	table += "INT IDENTITY(1,1) NOT NULL";
	table += ", layer_id ";
	table += "INT ";
	table += ", attr_table ";
	table += "VARCHAR(255) NOT NULL ";
	table += ", unique_id ";
	table += "VARCHAR(255) ";
	table += ", attr_link ";
	table += "VARCHAR(255) ";
	table += ", attr_initial_time ";
	table += "VARCHAR(255) ";
	table += ", attr_final_time ";
	table += "VARCHAR(255) ";
	table += ", attr_time_unit ";
	table += "NUMERIC(20) ";
	table += ", attr_table_type ";
	table += "NUMERIC(20) ";
	table += ", user_name ";
	table += "VARCHAR(255) ";
	table += ", initial_time ";
	table += "DATETIME ";
	table += ", final_time ";
	table += "DATETIME ";

	table += ", PRIMARY KEY (table_id))";

	if(!execute(table))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_layer_table!";
		return false;
	}

	return true;
}

bool 
TeSqlServer::createTablesRelationTable()
{

	string table;
	table = "CREATE TABLE te_tables_relation (";
	table += "relation_id ";
	table += "INT IDENTITY(1,1) NOT NULL";
	table += ", related_table_id ";
	table += "INT NOT NULL ";
	table += ", related_attr ";
	table += "VARCHAR(255) NOT NULL";
	table += ", external_table_name ";
	table += "VARCHAR(255) NOT NULL";
	table += ", external_attr ";
	table += "VARCHAR(255) NOT NULL ";

	table += ", PRIMARY KEY (relation_id))";

	if(!execute(table))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_tables_relation!";
		return false;
	}

	return true;

}

bool 
TeSqlServer::createRepresentationTable ()
{
	string table;
	table = "CREATE TABLE te_representation (";
	table += " repres_id ";
	table += "INT IDENTITY(1,1) NOT NULL";
	table += ", layer_id ";
	table += "INT NOT NULL";
	table += ", geom_type ";
	table += "SMALLINT NOT NULL";
	table += ", geom_table ";
	table += "VARCHAR(255) NOT NULL";
	table += ", description "; 
	table += "VARCHAR(255) ";
	table += ", lower_x ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL";
	table += ", lower_y ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL";
	table += ", upper_x ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL";
	table += ", upper_y ";
	table += "DOUBLE PRECISION DEFAULT 0.0000 NOT NULL";
	table += ", res_x ";
	table += "DOUBLE PRECISION ";
	table += ", res_y ";
	table += "DOUBLE PRECISION ";
	table += ", num_cols ";
	table += "NUMERIC(20)";
	table += ", num_rows ";
	table += "NUMERIC(20)";
	table += ", initial_time ";
	table += "DATETIME";
	table += ", final_time ";
	table += "DATETIME";
	
	table += ", PRIMARY KEY (repres_id))";
 
	if(!execute(table))
	{	
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_representation!";
		return false;
	}

	return true;
}

bool
TeSqlServer::createRasterMetadataTable (const string& tableName)
{
	
	if(tableExist(tableName))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
	
	string table;
	table = "CREATE TABLE "+ tableName +" (";
	table += " geom_id ";
	table += "INT NOT NULL";
	table += ", band_id ";
	table += "NUMERIC(32) NOT NULL";
	table += ", min_value ";
	table += "DOUBLE PRECISION DEFAULT 0.0 ";
	table += ", max_value ";
	table += "DOUBLE PRECISION DEFAULT 0.0 ";
	table += ", num_bits ";
	table += "NUMERIC(32)";
	table += ", data_type ";
	table += "NUMERIC(32)";
	table += ", photometric_type ";
	table += "NUMERIC(32)";
	table += ", compression_type ";
	table += "NUMERIC(32)";
	table += ", [dummy] ";
	table += "DOUBLE PRECISION ";
	
	table += ", PRIMARY KEY (geom_id, band_id))";

	if(!execute(table))
	{	
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table "+ tableName +"!";
		return false;
	}
	
	return true;
}


bool
TeSqlServer::createLUTTable(const string& table)
{
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
	
	string tab;
	tab = "CREATE TABLE " + table + " ( ";
	tab += "index_id ";
	tab += "INT NOT NULL";
	tab += ", r_val ";
	tab += "TINYINT NOT NULL ";
	tab += ", g_val ";
	tab += "TINYINT NOT NULL ";
	tab += ", b_val ";
	tab += "TINYINT NOT NULL ";
	
	tab += ", PRIMARY KEY (index_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table ;
		return false;
	}

	
	return true;
}


bool 
TeSqlServer::createViewTable()
{
	string table;
	table = "CREATE TABLE te_view ( ";
	table += "view_id ";
	table += "INT IDENTITY(1,1) NOT NULL";
	table += ", projection_id ";
	table += "INT NOT NULL";
	table += " CONSTRAINT view_proj UNIQUE ";
	table += ", name ";
	table += "VARCHAR(255) NOT NULL";
	table += ", user_name ";
	table += "VARCHAR(255) ";
	table += ", visibility ";
	table += "BIT NULL ";

	table += ", PRIMARY KEY(view_id))";
	
	if(!execute(table))
	{	
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_view!";
		return false;
	}
	
	//creates index
	if (!createIndex ("te_view", "view_index", "name, user_name"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table te_view!";
		return false;
	}
	
	return true;
}

bool 
TeSqlServer::createThemeTable ()
{
	string table;
	table = "CREATE TABLE te_theme (";
	table += "theme_id ";
	table += "INT IDENTITY(1,1) NOT NULL";
	table += ", layer_id ";
	table += "INT NULL";
	table += ", view_id ";
	table += "INT NULL";
	table += ", name ";
	table += "VARCHAR(255) NOT NULL";
	table += ", parent_id ";
	table += "INT ";
	table += ", priority ";
	table += "NUMERIC(20) ";	
	table += ", node_type ";
	table += "NUMERIC(20) NOT NULL";	
		
	table += ", min_scale ";
	table += "NUMERIC(20,12)";
	table += ", max_scale ";
	table += "NUMERIC(20,12)";

	table += ", generate_attribute_where ";
	table += "VARCHAR(500)";
	table += ", generate_spatial_where ";
	table += "VARCHAR(500)";
	table += ", generate_temporal_where ";
	table += "VARCHAR(500)";
	
	table += ", collection_table ";
	table += "VARCHAR(255)";
	
	table += ", visible_rep ";
	table += "INT NULL";
	table += ", enable_visibility ";
	table += "INT NULL";

	table += ", PRIMARY KEY (theme_id))";
	
	if(!execute(table))
	{	
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_theme!";
		return false;
	}

	return true;
}

bool 
TeSqlServer::createThemeTablesTable()
{
	string table = "CREATE TABLE te_theme_table  (";
	table += " theme_table_id ";
	table += "INT IDENTITY(1,1) NOT NULL";
	table += ", theme_id ";
	table += " INT NOT NULL ";
	table += ", table_id ";
	table += " INT NOT NULL ";
	table += ", relation_id ";
	table += " INT";
	table += ", table_order ";
	table += " NUMERIC(20) ";
	table += ", PRIMARY KEY (theme_table_id) ";
	table += ")";
	
	if(!execute(table))
	{	
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_theme_table!";
		return false;
	}
	
	//creates index
	if (!createIndex ("te_theme_table", "theme_table_index", "theme_id"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table te_theme_table!";
		return false;
	}

	return true;
}

bool
TeSqlServer::createGroupingTable()
{
	string table;
	table = "CREATE TABLE te_grouping (";
	table += "theme_id ";
	table += "INT NOT NULL";
	table += ", grouping_number ";
	table += "NUMERIC(20)";
	table += ", grouping_attr ";
	table += "VARCHAR(255)";
	table += ", grouping_attr_type ";
	table += "NUMERIC(20)";
	table += ", grouping_mode ";
	table += "NUMERIC(20)";
	table += ", grouping_norm_attr ";
	table += "VARCHAR(255)";
	table += ", grouping_std_dev ";
	table += "DOUBLE PRECISION DEFAULT 0.0";
	table += ", grouping_precision ";
	table += "NUMERIC(20,12)";
	table += ", grouping_function ";
	table += "VARCHAR(20)";
	table += ", grouping_chronon ";
	table += "NUMERIC(20)";
	table += ", PRIMARY KEY (theme_id))";
	
	if(!execute(table))
	{	
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_grouping!";
		return false;
	}
	return true;
}


bool 
TeSqlServer::createLegendTable ()
{
	string table;
	table = "CREATE TABLE te_legend (";
	table += "  legend_id ";
	table += "INT IDENTITY(1,1) NOT NULL";
	table += ", theme_id ";
	table += "INT NOT NULL";
	table += ", group_id ";
	table += "INT";
	table += ", num_objs ";
	table += "NUMERIC(20)";
	table += ", lower_value ";
	table += "VARCHAR(255)";
	table += ", upper_value ";
	table += "VARCHAR(255)";
	table += ", label ";
	table += "VARCHAR(255)";

	table += ", PRIMARY KEY (legend_id))";

	if(!execute(table))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_legend!";
		return false;
	}
	
	//creates index
	if (!createIndex ("te_legend", "legend_index", "theme_id"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table te_legend!";
		return false;
	}

	return true;
}

bool 
TeSqlServer::createVisualTable ()
{
	string table;
	table = "CREATE TABLE te_visual (";
	table += "  legend_id ";
	table += "INT NOT NULL";
	table += ", geom_type ";
	table += "SMALLINT NOT NULL";

	table += ", lib_name ";
	table += "VARCHAR(255) ";
	table += ", symb_id ";
	table += "NUMERIC(20) ";
	table += ", red ";
	table += "TINYINT ";
	table += ", green ";
	table += "TINYINT ";
	table += ", blue ";
	table += "TINYINT ";
	table += ", transparency ";
	table += "TINYINT ";
	table += ", width ";
	table += "TINYINT ";
	
	table += ", contour_lib_name ";
	table += "VARCHAR(255) ";
	table += ", contour_symb_id ";
	table += "NUMERIC(20) ";
	table += ", contour_red ";
	table += "TINYINT ";
	table += ", contour_green ";
	table += "TINYINT ";
	table += ", contour_blue ";
	table += "TINYINT ";
	table += ", contour_transp ";
	table += "NUMERIC(20) ";
	table += ", contour_width ";
	table += "TINYINT ";

	table += ", size_value ";
	table += "NUMERIC(20) ";
	table += ", pt_angle ";
	table += "NUMERIC(20) ";
	
	table += ", family ";
	table += "VARCHAR(100)";
	table += ", bold ";
	table += "BIT ";
	table += ", italic ";
	table += "BIT ";
	table += ", alignment_vert ";
	table += " DOUBLE PRECISION ";
	table += ", alignment_horiz ";
	table += " DOUBLE PRECISION ";
	table += ", tab_size ";
	table += "NUMERIC(20) ";
	table += ", line_space ";
	table += "NUMERIC(20) ";
	table += ", fixed_size ";
	table += "NUMERIC(20) ";

	table += ", PRIMARY KEY (legend_id, geom_type))";

	if(!execute(table))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_visual!";
		return false;
	}
	
	return true;
}

bool 
TeSqlServer::createVisualRasterTable()
{
	string create = "CREATE TABLE te_visual_raster (";
	create += " theme_id		INT NOT NULL,";
	create += " band_in			NUMERIC(20) NOT NULL,";
	create += " band_out		NUMERIC(20), ";
	create += " transf_type		NUMERIC(20), ";
	create += " param1			DOUBLE PRECISION,";
	create += " param2			DOUBLE PRECISION,";
	create += " lut_table		VARCHAR(255), ";
	create += " PRIMARY KEY (theme_id, band_in))";
	
	if(!execute(create))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_visual_raster!";
		return false;
	}
	
	return true;
}

bool 
TeSqlServer::createCollectionTable(const string& tableName)
{

	if(tableExist(tableName))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
	
	string table = "CREATE TABLE "+ tableName +" (";
	table += " c_object_id ";
	table += "VARCHAR(255) NOT NULL";
	table += ", c_legend_id  ";
	table += "INT ";
	table += ", label_x ";
	table += " DOUBLE PRECISION ";
	table += ", label_y ";
	table += " DOUBLE PRECISION ";
	table += ", c_legend_own ";
	table += "NUMERIC(20) ";
	table += ", c_object_status ";
	table += "NUMERIC(20) ";
	
	table += ", PRIMARY KEY (c_object_id))";

	if(!execute(table))
	{	
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + tableName + "!";
		return false;
	}
		
	return true;
}


bool 
TeSqlServer::createCellGeometry (const string& table)
{
	
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
	
	string tab;
	tab = "CREATE TABLE " + table + "(";
	tab += "geom_id ";
	tab += "INT IDENTITY(1,1) NOT NULL";
	tab += ", object_id ";
	tab += "VARCHAR(255) NULL";
	tab += ", col_number ";
	tab += "NUMERIC(20) NOT NULL";
	tab += ", row_number ";
	tab += "NUMERIC(20) NOT NULL";
	tab += ", lower_x ";
	tab += "DOUBLE PRECISION";
	tab += ", lower_y ";
	tab += "DOUBLE PRECISION";
	tab += ", upper_x ";
	tab += "DOUBLE PRECISION";
	tab += ", upper_y ";
	tab += "DOUBLE PRECISION";
	tab += ", PRIMARY KEY (geom_id))";

	if (!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table + " !" ;
		return false;
	}

	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}

	if(!createIndex(table, table + "box_idx", "lower_x, lower_y, upper_x, upper_y"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}


	if(!createIndex(table, table + "rc_idx", "row_number, col_number"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}

	
	return true;
}

bool 
TeSqlServer::createTextGeometry (const string& table)
{
	
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
	
	string tab;
	tab = "CREATE TABLE " + table +"(";
	tab += "geom_id ";
	tab += "INT IDENTITY(1,1) NOT NULL";
	tab += ", object_id ";
	tab += "VARCHAR(255) ";
	tab += ", x ";
	tab += "DOUBLE PRECISION NULL";
	tab += ", y ";
	tab += "DOUBLE PRECISION NULL";
	tab += ", text_value ";
	tab += "VARCHAR(255) NULL";
	tab += ", angle ";
	tab += "DOUBLE PRECISION NULL";
	tab += ", height ";
	tab += "DOUBLE PRECISION NULL";
	tab += ", alignment_vert ";
	tab += "DOUBLE PRECISION NULL";
	tab += ", alignment_horiz ";
	tab += "DOUBLE PRECISION NULL";
	
	tab += ", PRIMARY KEY (geom_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table  " + table + " !";
		return false;
	}

	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}

	if(!createIndex(table, table + "xy_idx", "x,y"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}
	
	
	return true;
}


bool 
TeSqlServer::createArcGeometry (const string& table)
{
	
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}

	string tab;
	tab = "CREATE TABLE " + table +"(";
	tab += "geom_id ";
	tab += "INT IDENTITY(1,1) NOT NULL";
	tab += ", object_id ";
	tab += "VARCHAR(255) NULL";
	tab += ", from_node ";
	tab += "NUMERIC(20) NOT NULL";
	tab += ", to_node ";
	tab += "NUMERIC(20) NOT NULL";
	
	tab += ", PRIMARY KEY (geom_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table ;
		return false;
	}

	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}

	if(!createIndex(table, table + "fn_idx", "from_node"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}

	if(!createIndex(table, table + "tn_idx", "to_node"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}
		
	
	return true;
}

bool 
TeSqlServer::createNodeGeometry (const string& table)
{
	
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
		
	string tab;
	tab = "CREATE TABLE " + table + " (";
	tab += "geom_id ";
	tab += "INT IDENTITY(1,1) NOT NULL";
	tab += ", object_id ";
	tab += "VARCHAR(255) NULL";
	tab += ", x ";
	tab += "DOUBLE PRECISION NOT NULL";
	tab += ", y ";
	tab += "DOUBLE PRECISION NOT NULL";
	tab += ", PRIMARY KEY (geom_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table  " + table + " !";
		return false;
	}
	
	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}

	if(!createIndex(table, table + "xy_idx", "x,y"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}
		
	return true;
}

bool 
TeSqlServer::createRasterGeometry (const string& table)
{
	
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}

	string create = "CREATE TABLE " + table + " (";
	create += "geom_id		INT IDENTITY(1,1) NOT NULL,";
	create += "object_id	VARCHAR(255) NOT NULL,";
	create += "raster_table VARCHAR(255) NOT NULL,";
	create += "lut_table	VARCHAR(255) ,";
	create += "res_x		DOUBLE PRECISION DEFAULT 0.0,";
	create += "res_y		DOUBLE PRECISION DEFAULT 0.0,";
	create += "num_bands	NUMERIC(20),";
	create += "num_cols		NUMERIC(20),";
	create += "num_rows		NUMERIC(20),";
	create += "block_height NUMERIC(20),";
	create += "block_width	NUMERIC(20),";
	create += "lower_x		DOUBLE PRECISION DEFAULT 0.0,";
	create += "lower_y		DOUBLE PRECISION DEFAULT 0.0,";
	create += "upper_x		DOUBLE PRECISION DEFAULT 0.0,";
	create += "upper_y		DOUBLE PRECISION DEFAULT 0.0,";
	create += "tiling_type  NUMERIC(20),";
	create += "PRIMARY KEY (geom_id))";
	
	if(!execute (create))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table + " !" ;
		return false;
	}
	
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}
		
	return true;
}

bool 
TeSqlServer::createRasterTable (const string& tableName)
{
	string tab;
	tab = "CREATE TABLE "+ tableName + " ( ";
	tab += "block_id ";
	tab += " VARCHAR(50) NOT NULL";
	tab += ", band_id ";
	tab += " TINYINT NOT NULL";
	tab += ", resolution_factor ";
	tab += " TINYINT ";
	tab += ", subband ";
	tab += " TINYINT ";
	tab += ", lower_x ";
	tab += " DOUBLE PRECISION DEFAULT 0.0 ";
	tab += ", lower_y ";
	tab += " DOUBLE PRECISION DEFAULT 0.0 ";
	tab += ", upper_x ";
	tab += " DOUBLE PRECISION DEFAULT 0.0 ";
	tab += ", upper_y ";
	tab += " DOUBLE PRECISION DEFAULT 0.0 ";
	tab += ", block_size ";
	tab += " NUMERIC(20) ";
	tab += ", spatial_data ";
	tab += " IMAGE ";
	
	tab += ", PRIMARY KEY (block_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + tableName ;
		return false;
	}
	
	if(!createIndex(tableName, tableName + "box_resfac_idx", "lower_x, lower_y, upper_x, upper_y, resolution_factor, subband"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + tableName + " !" ;
		return false;
	}

	if(!createIndex(tableName, tableName + "band_id_idx", "band_id"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + tableName + " !" ;
		return false;
	}

	return true;
}

bool
TeSqlServer::deleteLayer(int layerId)
{
	//Delete attributes tables
	if(!deleteLayerTable(layerId))
		return false;

	TeSqlServerPortal* portal = (TeSqlServerPortal*) this->getPortal();
	if (!portal)
		return false;

	string geomTable;
	string sql = "SELECT projection_id FROM te_layer WHERE layer_id = ";
	sql += Te2String(layerId);

	if (!portal->query(sql))
	{	
		delete portal;
		return false;
	}

	if (!portal->fetchRow())
	{
		delete portal;
		return false;
	}
	string projId = portal->getData("projection_id");
	portal->freeResult();

	// Get all representations that are associated to this layer
	sql = "SELECT * FROM te_representation WHERE layer_id = "+ Te2String(layerId);
	if (!portal->query (sql))
	{
		delete portal;
		return false;
	}

	while (portal->fetchRow())
	{	// Delete the geometry tables
		geomTable = portal->getData("geom_table");
	
		// Delete lut table
		TeGeomRep rep = TeGeomRep(atoi(portal->getData("geom_type")));
		if (rep == TeRASTER)
		{
			TeSqlServerPortal* portal2 = (TeSqlServerPortal*)this->getPortal();
			sql = "SELECT lut_table, raster_table FROM " + geomTable;
			string tabName;
			if (!portal2->query (sql))
			{
				delete portal2;
				continue;
			}

			while (portal2->fetchRow())
			{
				// remove lut table
				tabName = portal2->getData(0);
				if (!tabName.empty() && this->tableExist(tabName))
				{
					sql = "DROP TABLE " + tabName;
					this->execute(sql);

				}
				// remove raster table
				tabName = portal2->getData(1);
				if (!tabName.empty() && this->tableExist(tabName))
				{
					sql = "DROP TABLE " + tabName;
					this->execute(sql);
				}
			}
			delete portal2;
			// remove raster metadata table
			tabName = geomTable + "_metadata";
			if (!tabName.empty() && this->tableExist(tabName))
			{
				sql = "DROP TABLE " + tabName;
				this->execute(sql);
			}
		}
		if (this->tableExist(geomTable))
		{
			sql = "DROP TABLE " + geomTable;
			if (!this->execute(sql) )
			{
				delete portal;
				return false;
			}
			
		}
	}

	portal->freeResult();
	if (existRelation("te_representation","fk_rep_layer_id") != TeRICascadeDeletion)
	{
		// Delete entries into representations table
		sql = "DELETE FROM te_representation WHERE layer_id = " +Te2String(layerId);
		if (!this->execute(sql) )
		{
			delete portal;
			return false;
		}
	}

	
	// delete layer themes
	sql = "SELECT theme_id FROM te_theme WHERE layer_id=" + Te2String(layerId);
	if (!portal->query (sql))
	{
		delete portal;
		return false;
	}
	
	int themeId;
	while (portal->fetchRow())
	{	
		themeId = atoi(portal->getData("theme_id"));
		this->deleteTheme(themeId);
	}


	
	sql = "DELETE FROM te_layer WHERE layer_id=" + Te2String(layerId);
	if (!this->execute(sql))
	{
		delete portal;
		return false;
	}

	// delete layer projection
	sql = "DELETE FROM te_projection WHERE projection_id = "+ projId;
	if (!this->execute(sql))
	{	
		delete portal;
		return false;
	}

	// remove all the items themes associated to the layer to be removed
	TeThemeMap::iterator it;
	for (it = themeMap_.begin(); it != themeMap_.end(); ++it)
	{
		TeTheme *theme = it->second;
		if (theme->layer()->id() == layerId)
		{
			themeMap_.erase(theme->id());
			delete theme;
		}
	}

	// Delete layer and its entry in the layer map
	TeLayer* layer = layerMap_[layerId];
	layerMap_.erase(layerId);
	delete layer;

	delete portal;
	return true;
}



bool
TeSqlServer::deleteTheme(int themeId)
{
   string sql;
   // delete the collection table associated to this theme
   TeDatabasePortal* portal = this->getPortal();
   if(!portal)
       return false;

   sql = "SELECT collection_table FROM te_theme WHERE theme_id = " + Te2String(themeId);
   if(!portal->query(sql) || !portal->fetchRow())
   {          
	   delete portal;
       return false;
   }
    string colTab = portal->getData("collection_table");
   //delete collection table
   if (this->tableExist(colTab))      
   {
       sql = "DROP TABLE " + colTab;
       if (!this->execute(sql) )
       {   
		   delete portal;
           return false;
       }
   }
   //delete auxiliar collection table
   if (this->tableExist(colTab +"_aux"))      
   {
       sql = "DROP TABLE " +colTab +"_aux";
       if (!this->execute(sql) )
       {              
		   delete portal;
           return false;
       }
   }
   portal->freeResult();

   //delete the visual associated to this theme
   if (existRelation("te_visual","fk_visual_legend_id") != TeRICascadeDeletion)
   {
       sql = "SELECT legend_id FROM te_legend WHERE theme_id = " + Te2String(themeId);
       if(!portal->query(sql))
       {              
		   delete portal;
           return false;
       }
       string wherec;
       int c = 0;
       while (portal->fetchRow())
       {
           if (c)
               wherec += ",";
           c++;
           wherec += portal->getData(0);
       }
       portal->freeResult();
       if (!wherec.empty()) 
       {
           sql = "DELETE FROM te_visual WHERE legend_id IN (" + wherec + ")";
           if (!this->execute(sql))
           {                  
			   delete portal;
               return false;
           }
       }
   }

   //delete all visuals of raster associated to this theme
   if (existRelation("te_visual_raster","fk_visrast_theme_id") != TeRICascadeDeletion)
   {
       sql = "DELETE FROM te_visual_raster WHERE theme_id =" + Te2String(themeId);
       if (!this->execute (sql))
       {              
		   delete portal;
           return false;
       }
   }
     //delete all legends associated to this theme
   if (existRelation("te_legend","fk_legend_theme_id") != TeRICascadeDeletion)
   {
       sql = "DELETE FROM te_legend WHERE theme_id =" + Te2String(themeId);
       if (!this->execute (sql))
       {              
		   delete portal;
           return false;
       }
   }
     //select the view of this theme
   sql = "SELECT view_id FROM te_theme WHERE theme_id = " + Te2String(themeId);
   portal->freeResult();
   if(!portal->query(sql) ||!portal->fetchRow())
   {          
	   delete portal;
       return false;
   }
 
   int viewId = portal->getInt("view_id");
   TeView* view = viewMap_[viewId];
   sql = "SELECT name FROM te_theme WHERE theme_id = " + Te2String(themeId);
   portal->freeResult();
   if (!portal->query(sql) || !portal->fetchRow())
   {          
	   delete portal;
       return false;
   }
  
   TeTheme* tema = NULL;
   if (view != 0)
       tema = (TeTheme*)view->remove(themeId);
   if (tema != 0) 
   {
       unsigned int i;
       TeLegendEntryVector& legendVector = tema->legend();
       for (i = 0; i < legendVector.size(); ++i)
           legendMap_.erase(legendVector[i].id());
       themeMap_.erase(themeId);
       delete tema;
   }
   delete portal;

   //delete the tables associated to this theme
   if (existRelation("te_theme_table","fk_thmtable_theme_id") != TeRICascadeDeletion)
   {
       sql = "DELETE FROM te_theme_table WHERE theme_id =" + Te2String(themeId);
       if (!this->execute (sql))
           return false;
   }
   //delete the grouping
   if (existRelation("te_grouping","fk_group_theme_id")  != TeRICascadeDeletion)
   {
       sql = "DELETE FROM te_grouping WHERE theme_id =" + Te2String(themeId);
       if (!this->execute (sql))
           return false;
   }

   if (existRelation("te_theme_application","fk_themeAppl_theme_id") != TeRICascadeDeletion)
   {
       sql="DELETE FROM te_theme_application where theme_id=" + Te2String(themeId);
       if(!this->execute(sql))
           return false;
   }

   // delete the theme
   sql = "DELETE FROM te_theme WHERE theme_id = " + Te2String(themeId);
   if (!this->execute (sql))
       return false;

   return true;
} 

bool 
TeSqlServer::deleteView (int viewId)
{
	TeDatabasePortal* portal = this->getPortal();

	// view projection should be deleted manually
	string sql =  "SELECT projection_id FROM te_view WHERE view_id=" + Te2String(viewId);
	portal->freeResult();
	if (!portal->query(sql))
	{
		delete portal;
		return false;
	}
	if (!portal->fetchRow())
	{
		delete portal;
		return false;
	}
	string projId = portal->getData("projection_id");
	portal->freeResult();
	
	// delete themes belonging to this view 
	sql = "SELECT theme_id FROM te_theme WHERE view_id=" + Te2String(viewId);
	if (!portal->query(sql))
	{
		delete portal;
		return false;
	}
	while (portal->fetchRow())
	{
		int id = atoi(portal->getData(0));
		if(deleteTheme(id) == false)
		{	
			delete portal;
			return false;
		}
	}

	// delete view
	sql = "DELETE FROM te_view WHERE view_id = " + Te2String(viewId);
	if (!this->execute (sql))
	{
		delete portal;
		return false;
	}

	sql = "DELETE FROM te_projection WHERE  projection_id = " + projId;
	if (!this->execute (sql))
	{
		delete portal;
		return false;
	}

	// Delete the view and its themes
	TeView* view = viewMap_[viewId];
	viewMap_.erase(viewId);
	delete view;
	delete portal;
	return true;
}

bool 
TeSqlServer::createPolygonGeometry (const string& table)
{
	
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
	
	string tab;
	tab = "CREATE TABLE " + table + " ( ";
	tab += "geom_id ";
	tab += "INT IDENTITY(1,1) NOT NULL";
	tab += ", object_id ";
	tab += "VARCHAR(255) NULL";
	tab += ", num_coords ";
	tab += "NUMERIC(20) NOT NULL ";
	tab += ", num_holes ";
	tab += "NUMERIC(20) NOT NULL ";
	tab += ", parent_id ";
	tab += "INT NOT NULL ";
	tab += ", lower_x ";
	tab += "DOUBLE PRECISION NOT NULL ";
	tab += ", lower_y ";
	tab += "DOUBLE PRECISION NOT NULL ";
	tab += ", upper_x ";
	tab += "DOUBLE PRECISION NOT NULL ";
	tab += ", upper_y ";
	tab += "DOUBLE PRECISION NOT NULL ";
	tab += ", ext_max ";
	tab += "DOUBLE PRECISION NOT NULL ";
	tab += ", spatial_data ";
	tab += "IMAGE ";	
	tab += ", PRIMARY KEY (geom_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table ;
		return false;
	}

	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}

	if(!createIndex(table, table + "parent_idx", "parent_id"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}

	if(!createIndex(table, table + "extm_idx", "ext_max"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}

	if(!createIndex(table, table + "box_idx", "lower_x, lower_y, upper_x, upper_y"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}

	return true;
}

bool 
TeSqlServer::createLineGeometry (const string& table)
{
	
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}

	string tab;
	tab = "CREATE TABLE " + table + " ( ";
	tab += "geom_id ";
	tab += "INT IDENTITY(1,1) NOT NULL";
	tab += ", object_id ";
	tab += "VARCHAR(255) NULL";
	tab += ", num_coords ";	
	tab += "NUMERIC(20) NOT NULL ";
	tab += ", lower_x ";
	tab += "DOUBLE PRECISION NOT NULL ";
	tab += ", lower_y ";
	tab += "DOUBLE PRECISION NOT NULL ";
	tab += ", upper_x ";
	tab += "DOUBLE PRECISION NOT NULL ";
	tab += ", upper_y ";
	tab += "DOUBLE PRECISION NOT NULL ";
	tab += ", ext_max ";
	tab += "DOUBLE PRECISION NOT NULL ";
	tab += ", spatial_data ";
	tab += "IMAGE ";	
	tab += ", PRIMARY KEY (geom_id))";


	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table + " !" ;
		return false;
	}

	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}

	if(!createIndex(table, table + "box_idx", "lower_x, lower_y, upper_x, upper_y"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}
			
	
	return true;
}

bool 
TeSqlServer::createPointGeometry (const string& table)
{
	
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}

	string tab;
	tab = "CREATE TABLE " + table + " ( ";
	tab += "geom_id ";
	tab += "INT IDENTITY(1,1) NOT NULL";
	tab += ", object_id ";
	tab += "VARCHAR(255) NULL";
	tab += ", x ";
	tab += "DOUBLE PRECISION ";
	tab += ", y ";	
	tab += "DOUBLE PRECISION ";
	tab += ", PRIMARY KEY (geom_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table + " !" ;
		return false;
	}	
		
	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}

	if(!createIndex(table, table + "xy_idx", "x,y"))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating index to table " + table + " !" ;
		return false;
	}
		
	
	return true;
}

bool 
TeSqlServer::insertTheme (TeTheme *theme)
{

	ADODB::_RecordsetPtr recset_;
	recset_.CreateInstance(__uuidof(ADODB::Recordset));
	try
	{
		recset_->Open("te_theme",
			_variant_t((IDispatch*)connection_,true),
			ADODB::adOpenKeyset,
			ADODB::adLockOptimistic,
			ADODB::adCmdTable);
		recset_->AddNew();
		recset_->Fields->GetItem("layer_id")->Value = (_variant_t) ((long)theme->layerId());
		recset_->Fields->GetItem("view_id")->Value = (_variant_t) ((long)theme->view());
		recset_->Fields->GetItem("name")->Value = (_bstr_t) (theme->name().c_str ());
		recset_->Fields->GetItem("parent_id")->Value = (_variant_t) ((long)theme->parentId());
		recset_->Fields->GetItem("priority")->Value = (_variant_t) ((long)theme->priority());
		recset_->Fields->GetItem("node_type")->Value = (_variant_t) ((long)theme->type());
		recset_->Fields->GetItem("min_scale")->Value = (_variant_t) (theme->minScale());
		recset_->Fields->GetItem("max_scale")->Value = (_variant_t) (theme->maxScale());
		
		if(!theme->attributeRest().empty())
			recset_->Fields->GetItem("generate_attribute_where")->Value = (_bstr_t) (theme->attributeRest().c_str());
		if(!theme->spatialRest().empty())
			recset_->Fields->GetItem("generate_spatial_where")->Value = (_bstr_t) (theme->spatialRest().c_str());
		if(!theme->temporalRest().empty())
			recset_->Fields->GetItem("generate_temporal_where")->Value = (_bstr_t) (theme->temporalRest().c_str());
		if(!theme->collectionTable().empty())
			recset_->Fields->GetItem("collection_table")->Value = (_bstr_t) (theme->collectionTable().c_str());
		
		recset_->Fields->GetItem("visible_rep")->Value = (_variant_t) ((long)theme->visibleRep());
		recset_->Fields->GetItem("enable_visibility")->Value = (_variant_t) ((long)theme->visibility());
		
		recset_->Update();
		_variant_t newID;
		newID = recset_->GetCollect("theme_id");
		theme->id(newID.intVal);

		recset_->Update();
		recset_->Close();

		if(theme->collectionTable().empty())
			{
				theme->collectionTable("te_collection_" + Te2String(theme->id()));
				string strSQL="update te_theme set collection_table='te_collection_" + Te2String(theme->id()) + "' where ";
				strSQL+= " theme_id=" + Te2String(theme->id());
				execute(strSQL);
			}
	}
	catch(_com_error &e)
	{
		errorMessage_ = e.Description();
		return false;
	}
	catch(...)
	{
		errorMessage_ = "Oppps !";
		return false;
	}
	
	bool status;

	// insert grouping
	int numSlices = 0;
	if(theme->grouping().groupMode_ != TeNoGrouping)
	{
		if(!insertGrouping (theme->id(), theme->grouping()))
			return false;
		numSlices = theme->grouping().groupNumSlices_;
	}

//	if(theme->rasterVisual()!=0)
//	{
//		if(!insertRasterVisual(theme->id(), theme->rasterVisual()))
//			return false;
//	}
		
	// insert legend
	theme->outOfCollectionLegend().group(-1); 
	theme->outOfCollectionLegend().theme(theme->id()); 
	status = insertLegend (&(theme->outOfCollectionLegend())); 
	if (!status)
		return status;

	theme->withoutDataConnectionLegend().group(-2); 
	theme->withoutDataConnectionLegend().theme(theme->id()); 
	status = insertLegend (&(theme->withoutDataConnectionLegend())); 
	if (!status)
		return status;

	theme->defaultLegend().group(-3); 
	theme->defaultLegend().theme(theme->id()); 
	status = insertLegend (&(theme->defaultLegend())); 
	if (!status)
		return status;

	theme->pointingLegend().group(-4); 
	theme->pointingLegend().theme(theme->id()); 
	status = insertLegend (&(theme->pointingLegend())); 
	if (!status)
		return status;

	theme->queryLegend().group(-5); 
	theme->queryLegend().theme(theme->id()); 
	status = insertLegend (&(theme->queryLegend())); 
	if (!status)
		return status;

	theme->queryAndPointingLegend().group(-6); 
	theme->queryAndPointingLegend().theme(theme->id()); 
	status = insertLegend (&(theme->queryAndPointingLegend())); 
	if (!status)
		return status;

	for (int i = 0; i < numSlices; i++)
	{
		theme->legend()[i].group(i);
		theme->legend()[i].theme(theme->id());
		status = insertLegend (&(theme->legend()[i]));
		if (!status)
			return status;
	}
	if (!status)
		return status;

	themeMap_[theme->id()] = theme;
	return updateThemeTable(theme);
}


bool 
TeSqlServer::generateLabelPositions (TeTheme *theme)
{
	string	geomTable, upd;
	string	collTable = theme->collectionTable();
	
	if((collTable.empty()) || (!tableExist(collTable)))
		return false;

	if( theme->layer()->hasGeometry(TeCELLS)    || 
		theme->layer()->hasGeometry(TePOLYGONS) ||
		theme->layer()->hasGeometry(TeLINES) )
	{
		geomTable = theme->layer()->tableName(TeCELLS);
		
		if(geomTable.empty())
		{
			geomTable = theme->layer()->tableName(TePOLYGONS);
			if(geomTable.empty())
				geomTable = theme->layer()->tableName(TeLINES);
		}

		upd= "UPDATE " + collTable + " SET ";
		upd += " label_x = (SELECT MAX(lower_x + (upper_x - lower_x)/2) ";
		upd += "FROM " + geomTable + " WHERE object_id = c_object_id), ";
		
		upd += " label_y = (SELECT MAX(lower_y + (upper_y - lower_y)/2) ";
		upd += "FROM " + geomTable + " WHERE object_id = c_object_id) ";
		upd += " WHERE label_x IS NULL OR label_y IS NULL";
	}
	else if(theme->layer()->hasGeometry(TePOINTS))
	{
		geomTable = theme->layer()->tableName(TePOINTS);
		
		upd= " UPDATE " + collTable + " SET ";
		upd += " label_x = (SELECT MAX(x) ";
		upd += " FROM " + geomTable + " p WHERE object_id = c_object_id), ";
		
		upd += " label_y = (SELECT MAX(y) ";
		upd += " FROM " + geomTable + " p WHERE object_id = c_object_id) ";
		upd += " WHERE label_x IS NULL OR label_y IS NULL";
	}

	if(!execute(upd))
		return false;

	return true;
}


bool 
TeSqlServer::insertPolygonSet (const string& table, TePolygonSet &ps)
{
	if (ps.empty())
		return false;

	ADODB::_RecordsetPtr recset_;
	recset_.CreateInstance(__uuidof(ADODB::Recordset));
	recset_->CursorLocation=ADODB::adUseServer;

	recset_->Open(_bstr_t(table.c_str()),
			_variant_t((IDispatch*)connection_,true),
			ADODB::adOpenKeyset,
			ADODB::adLockOptimistic,
			ADODB::adCmdTable);
	
	unsigned int i;
	unsigned int n = ps.size();
	try
	{
		// for each polygon in the polygon set
		for (i = 0; i < n; ++i)
		{
			unsigned int k;
			int parentId = 0;
			unsigned int numberHoles = ps[i].size()-1;
			// for each ring 
			for (k=0; k<ps[i].size(); ++k)
			{
				recset_->AddNew();
				recset_->Fields->GetItem("object_id")->Value = (_bstr_t) (ps[i].objectId().c_str());
				recset_->Fields->GetItem("num_coords")->Value = (_variant_t) ((long)ps[i][k].size());
				recset_->Fields->GetItem("num_holes")->Value = (_variant_t) ((long)numberHoles);
				recset_->Fields->GetItem("parent_id")->Value = (_variant_t) ((long)parentId);
				recset_->Fields->GetItem("lower_x")->Value = (_variant_t) (ps[i][k].box().lowerLeft().x_);
				recset_->Fields->GetItem("lower_y")->Value = (_variant_t) (ps[i][k].box().lowerLeft().y_);
				recset_->Fields->GetItem("upper_x")->Value = (_variant_t) (ps[i][k].box().upperRight().x_);
				recset_->Fields->GetItem("upper_y")->Value = (_variant_t) (ps[i][k].box().upperRight().y_);
				recset_->Fields->GetItem("ext_max")->Value = (_variant_t) (MAX(ps[i][k].box().width(),ps[i][k].box().height()));

				// create and save the BLOB
				VARIANT varPoints;
				if(!LinearRingToVariant(varPoints, ps[i][k]))
				{
					recset_->Close();
					return false;
				}

				recset_->Fields->Item["spatial_data"]->AppendChunk (varPoints);
				recset_->Update();
				SafeArrayDestroy (varPoints.parray);

				// retrieve the geometry id of the ring (given by the auto increment property)
				ps[i][k].geomId(recset_->GetCollect("geom_id").intVal);
 
				if (k==0)
				{
					parentId = ps[i][k].geomId(); // save the geometry id of the parent
					numberHoles=0;
				}
			}
		}
	}
	catch(_com_error &e)
	{
		errorMessage_ = e.Description();
		recset_->Close();
		return false;
	}
	catch(...)
	{
		errorMessage_ = "Oppps !";
		recset_->Close();
		return false;
	}
	recset_->Close();
	// update in the database the parent id of an outter ring 
	string strSQL="update " + table + " set parent_id=geom_id where parent_id=0";
	execute(strSQL);
	return true;
}


bool 
TeSqlServer::insertPolygon (const string& table, TePolygon &pol)
{
	
	ADODB::_RecordsetPtr recset_;
	recset_.CreateInstance(__uuidof(ADODB::Recordset));
	recset_->CursorLocation=ADODB::adUseServer;

	recset_->Open(_bstr_t(table.c_str()),
			_variant_t((IDispatch*)connection_,true),
			ADODB::adOpenKeyset,
			ADODB::adLockOptimistic,
			ADODB::adCmdTable);
	try
	{
		unsigned int k;
		int parentId = 0;
		unsigned int numberHoles = pol.size()-1;
		// for each ring
		for (k=0; k<pol.size(); ++k)
		{
			recset_->AddNew();
			recset_->Fields->GetItem("object_id")->Value = (_bstr_t) (pol.objectId().c_str());
			recset_->Fields->GetItem("num_coords")->Value = (_variant_t) ((long)pol[k].size());
			recset_->Fields->GetItem("num_holes")->Value = (_variant_t) ((long)numberHoles);
			recset_->Fields->GetItem("parent_id")->Value = (_variant_t) ((long)parentId);
			recset_->Fields->GetItem("lower_x")->Value = (_variant_t) (pol[k].box().lowerLeft().x_);
			recset_->Fields->GetItem("lower_y")->Value = (_variant_t) (pol[k].box().lowerLeft().y_);
			recset_->Fields->GetItem("upper_x")->Value = (_variant_t) (pol[k].box().upperRight().x_);
			recset_->Fields->GetItem("upper_y")->Value = (_variant_t) (pol[k].box().upperRight().y_);
			recset_->Fields->GetItem("ext_max")->Value = (_variant_t) (MAX(pol[k].box().width(),pol[k].box().height()));

			// create and save the BLOB
			VARIANT varPoints;
			if(!LinearRingToVariant(varPoints,pol[k]))
			{
				recset_->Close();
				return false;
			}

			recset_->Fields->Item["spatial_data"]->AppendChunk (varPoints);
			recset_->Update();
			SafeArrayDestroy (varPoints.parray);

			// retrieve the geometry id of the ring (given by the auto increment property)
			pol[k].geomId(recset_->GetCollect("geom_id").intVal);

			if (k==0)
			{
				parentId = pol[k].geomId();// save the geometry id of the parent
				numberHoles=0;
			}
		}
	}
	catch(_com_error &e)
	{
		errorMessage_ = e.Description();
		recset_->Close();
		return false;
	}
	catch(...)
	{
		errorMessage_ = "Oppps !";
		recset_->Close();
		return false;
	}
	recset_->Close();

	// update in the database the parent id of the outter rings 
	string strSQL="update " + table + " set parent_id=geom_id where parent_id=0";
	return (execute(strSQL));
}

string 
TeSqlServer::getSQLTime(TeTime& time)
{
	std::string dateStr = time.getDate(); //yyyy-mm-dd
	std::string timeStr = time.getTime();	
	string result = "'"+dateStr+" "+timeStr+"'";
	return result;
}


//----- TeSqlServerPortal methods ---

TeSqlServerPortal::TeSqlServerPortal ( TeDatabase*  pDatabase) 
{
	db_ = pDatabase;
	connection_ = ((TeSqlServer*)pDatabase)->connection_;
	recset_.CreateInstance(__uuidof(ADODB::Recordset));
}


TeSqlServerPortal::TeSqlServerPortal ()
{
	recset_.CreateInstance(__uuidof(ADODB::Recordset));
	curRow_ = 0;
}

TeSqlServerPortal::~TeSqlServerPortal ()
{
	freeResult();
}

TeTime
TeSqlServerPortal::getDate (const string& name)
{
	_variant_t value;

	TeTime t;
	try
	{
		value = recset_->GetCollect(name.c_str());
	}
	catch(_com_error &e)
	{
		string field = TeGetExtension(name.c_str());
		if (!field.empty())
		{
			try
			{
				value = recset_->GetCollect(field.c_str());
			}
			catch(_com_error &e)
			{
				errorMessage_ = e.Description();
				return t;
			}
			catch(...)
			{
				errorMessage_ = "Error getDate!";
				return t;
			}
		}
		else
		{
			errorMessage_ = e.Description();
			return t;
		}
	}
	catch(...)
	{
		errorMessage_ = "Oppps !";
		return t;
	}

	if (value.vt != VT_NULL)
	{
		bvalue_ = _bstr_t(value);
		t = TeTime(string((char*)bvalue_), TeSECOND, "DD/MM/YYYY HH24:MI:SS"); 
		return t;
	}
	else
		return t;
}


TeTime
TeSqlServerPortal::getDate (int i)
{
    _variant_t vtIndex;
	_variant_t value;

    vtIndex.vt = VT_I4;
	vtIndex.lVal = i;
	
	TeTime t;
	try
	{
		value = recset_->GetFields()->GetItem(vtIndex)->Value;
	}
	catch(_com_error &e)
	{
		errorMessage_ = e.Description();
		return t;
	}
	catch(...)
	{
		errorMessage_ = "Oppps !";
		return t;
	}
	if (value.vt != VT_NULL)
	{
		bvalue_ = _bstr_t(value);
		t = TeTime (string((char*)bvalue_), TeSECOND, "DD/MM/YYYY HH24:MI:SS"); 		
		return t;
	}
	else
		return t;
}


string
TeSqlServerPortal::getDateAsString(int i)
{
	TeTime t = this->getDate(i);
	string date = t.getDateTime ();

	if (!date.empty())
	{		string tval = " TO_DATE ('"+ date +"','DDsMMsYYYYsHHsmmsSS') ";
		return tval;
	}
	else
		return "";
}

string 
TeSqlServerPortal::getDateAsString(const string& s)
{
	TeTime t = this->getDate(s);
	string date = t.getDateTime ();

	if (!date.empty())
	{		string tval = " TO_DATE ('"+ date +"','DDsMMsYYYYsHHsmmsSS') ";
		return tval;
	}
	else
		return "";
}



